#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
from pathlib import Path
import shlex
from typing import Callable

from p1st.exec_print_transfer import execute_print_transfer_chunks
from p1st.common import _get_chunk_file, _get_remote_socket, _get_chunk_tmpl
from p1st.transfer_inform import transfer_inform


def main():
    args = parse_args()

    command_parts = (
        ['btrfs', 'send'],
        ['-p', str(args.parent)] if args.parent else [],
        ['--compress-data'] if args.compressed_data else [],
        [str(args.child)]
    )
    command = [x for xs in command_parts for x in xs]

    chunk_tmpl = _get_chunk_tmpl(args.child)
    target_chunk_tmpl = _get_chunk_tmpl(args.target_path)

    execute_print_transfer_chunks(
        command=command,
        chunk_file_tmpl=chunk_tmpl,
        chunk_transfer_fun=chunk_transfer_fun,
        chunk_transfer_args=(args.ssh_target,
                             args.target_path,
                             _get_chunk_file,
                             target_chunk_tmpl),
        chunk_size=args.chunk_size,
        get_chunk_file=_get_chunk_file,
    )


def parse_args():
    parser = argparse.ArgumentParser(prog='btrfs-send-chunks')

    parser.add_argument('-p',
                        help='Path to parent subvolume; forwarded to btrfs-send.',
                        dest='parent',
                        default=None,
                        type=Path,
                        metavar='PARENT_SUBVOLUME'
                        )

    parser.add_argument('--compressed-data',
                        help='Forwarded to btrfs-send.',
                        dest='compressed_data',
                        action='store_true',
                        default=False,
                        )

    parser.add_argument('--chunk-size',
                        help='Size in bytes; defaults to 64 MB.',
                        dest='chunk_size',
                        type=int,
                        default=64 * 1024 * 1024,
                        )

    parser.add_argument('child',
                        help='Path to child subvolume. Forwarded to btrfs-send.',
                        type=Path,
                        metavar='CHILD_SUBVOLUME'
                        )

    parser.add_argument('ssh_target',
                        help='Hostname of target computer; as configured in ~/.ssh/config.',
                        metavar='SSH_TARGET'
                        )

    parser.add_argument('target_path',
                        help='Path where the subvolume will be created on the remote ssh target.',
                        type=Path,
                        metavar='TARGET_PATH'
                        )

    args = parser.parse_args()

    # Make all paths absolute.
    if args.parent:
        args.parent = args.parent.absolute()
    args.child = args.child.absolute()
    if args.target_path:
        args.target_path = args.target_path.absolute()

    return args


def chunk_transfer_fun(chunk_file: Path, ct: int, eof: bool,
                       ssh_target: str,
                       target_path: Path,
                       get_chunk_file: Callable[[Path, int], Path],
                       target_chunk_tmpl: Path,
                       ):
    target_chunk = get_chunk_file(target_chunk_tmpl, ct)
    rsync_cmd = ['rsync', str(chunk_file), f'{ssh_target}:{str(target_chunk)}']

    message = 'EOF' if eof else 'OK'
    target_socket = _get_remote_socket(target_path)
    inform_cmd = ['ssh', ssh_target, f'echo {message} | nc -U {shlex.quote(str(target_socket))}']

    transfer_inform(
        rsync_cmd=rsync_cmd,
        inform_cmd=inform_cmd,
        user_input_file=chunk_file.parent.joinpath(f'{chunk_file.name}.SOCKET'),
    )


if __name__ == '__main__':
    main()
